/*
 * File:  CoverSimplier.cpp
 * Author:  mikolas
 * Created on:  12 Feb 2017 11:37:12
 * Copyright (C) 2017, Mikolas Janota
 */
#include<algorithm>
#include"CoverSimplier.h"
#include"minisat_auxiliary.h"
using SATSPC::LSet;
using SATSPC::lit_Undef;
using std::vector;
using std::endl;
AigLit CoverSimplier::run() {
//    vector<LSet*> positive, negative;
//    make_samples(positive, negative);
    const auto p = simplify(trues);
    const auto n = simplify(falses);
    const auto retv = (util.size(p) <= util.size(n)) ? p : factory.neg(n);
    return retv;
}

/*
void CoverSimplier::make_samples(vector<LSet*>& positive,vector<LSet*>& negative) {
    for (auto core: cores) {
        const Lit tl = get_vlit(target_variable,*core);
        if (tl == lit_Undef) continue;
        auto& ts = sign(tl) ? negative : positive;
        ts.push_back(new LSet());
        for (auto v: domain) {
            const Lit dl = get_vlit(v, *core);
            if (dl != lit_Undef) ts.back()->insert(dl);
        }
    }
}
*/

struct LSetSzCmp { inline bool operator () (const LSet* a,const LSet* b) const { return a->size() < b->size(); } };

//LSet* CoverSimplier::resolve(LSet& a, LSet& b, Lit pivot) {
LSet* resolve(LSet& a, LSet& b, Lit pivot) {
    bool found = false;
    auto* retv = new LSet();
    for (int i = 0; i < a.size(); ++i) {
        const Lit l = a[i];
        assert(var(l) != var(pivot) || l == pivot);
        if (l == pivot) found = true;
        else retv->insert(l);
    }
    if (!found) assert(0);
    found = false;
    const Lit npivot = ~pivot;
    for (int i = 0; i < b.size(); ++i) {
        const Lit l = b[i];
        assert(var(l) != var(pivot) || l == npivot);
        if (l == npivot) found = true;
        else retv->insert(l);
    }
    if (!found) assert(0);
    return retv;
}

bool good_ix(vector<LSet*>& avec, size_t ai) { return ai < avec.size() && avec[ai]; }

void annihilate(vector<LSet*>& avec, size_t ai) {
    assert(good_ix(avec, ai));
    delete avec[ai];
    avec[ai] = NULL;
}

void subsume_chk(vector<LSet*>& _avec, vector<LSet*>& _bvec, size_t _ai, size_t _bi, vector<LSet*>& new_terms) {
    assert(good_ix(_avec, _ai));
    assert(good_ix(_bvec, _bi));
    const bool sw = _avec[_ai]->size() > _bvec[_bi]->size();
    vector<LSet*>& avec = sw ? _bvec : _avec;
    vector<LSet*>& bvec = sw ? _avec : _bvec;
    const auto ai = sw ? _bi : _ai;
    const auto bi = sw ? _ai : _bi;

    LSet& a = *(avec[ai]);
    LSet& b = *(bvec[bi]);
    bool subsumed = true;
    bool selfsubsumed = true;
    Lit pivot = lit_Undef;
    for (int li = 0; (selfsubsumed || subsumed) && li < a.size(); ++li) {
        const Lit al = a[li];
        const Lit bl = get_vlit(var(al), b);
        if (bl == lit_Undef) {
            subsumed = selfsubsumed = false;
        } else if (bl == al) {
            // nop
        } else {
            assert (bl == ~al);
            subsumed = false;
            if (pivot != lit_Undef) selfsubsumed = false;
            else pivot = al;
        }
    }
    if (subsumed) {
        annihilate(bvec, bi);
    } else if (selfsubsumed) {
        new_terms.push_back(resolve(a, b, pivot));
        annihilate(bvec, bi);
    }
}

void shake_down(vector<LSet*>& vs) {
    size_t i = 0;
    while (i < vs.size()) {
        if (vs[i] == NULL) {
            vs[i] = vs.back();
            vs.pop_back();
        } else {
            ++i;
        }
    }
}

void mv(vector<LSet*>& src, vector<LSet*>& dst) {
    for (auto* c : src) if (c) dst.push_back(c);
    src.clear();
}

AigLit CoverSimplier::simplify(vector<LSet*>& impls) { // TODO mem leak
    /*
    std::cerr<<"simplify impls"<<endl;
    for (auto i : impls) {
        std::cerr<<"impl:";
        for (int j = 0; j < i->size(); ++j) std::cerr<<" "<<(*i)[j];
        std::cerr<<endl;
    }
    */
    std::sort(impls.begin(), impls.end(), LSetSzCmp());
    vector<LSet*> tmp1, tmp2;
    vector<LSet*> *old = &tmp1;
    vector<LSet*> *active = &impls;
    vector<LSet*> *todo = &tmp2;

    while (!active->empty ()) {
        for (size_t i = 0; i < active->size(); ++i) {
            for (size_t j = i+1; (*active)[i] != NULL && j < active->size(); ++j) {
                if ((*active)[j] == NULL) continue;
                subsume_chk(*active, *active, i, j, *todo);
            }
        }

        for (size_t i = 0; i < active->size(); ++i) {
            for (size_t j = 0; (*active)[i] != NULL && j < old->size(); ++j) {
                if ((*old)[j] == NULL) continue;
                subsume_chk(*active, *old, i, j, *todo);
            }
        }
        shake_down(*old);
        mv(*active, *old);
        swap(active, todo);
    }
    const auto retv = disjoin(*old);
    //cleanup(*old);
    return retv;
}

AigLit CoverSimplier::disjoin(vector<LSet*>& impls) {
    vector<AigLit> tmp;
    vector<AigLit> tmp1;
    for (auto * c : impls) {
        assert(c);
        tmp1.clear();
        for (int i = 0;  i < c->size(); ++i) {
            const Lit l = (*c)[i];
            tmp1.push_back(factory.mk_var(var(l), sign(l)));
        }
        tmp.push_back(util.and_(tmp1, false));
    }
    return util.or_(tmp, false);
}
