/*
 * File:  dec_tree.h
 * Author:  mikolas
 * Created on:  10 Apr 2017 10:28:31
 * Copyright (C) 2017, Mikolas Janota
 */
#ifndef DEC_TREE_H_20858
#define DEC_TREE_H_20858
#include <unordered_map>
#include <vector>
#include "minisat_auxiliary.h"
#include "VarSet.h"
#include "aig.h"
#include "Play.h"
class DecTree {
    public:
        typedef Play T;
        DecTree(AigFactory& factory) : verb(0), factory(factory) {}
        virtual ~DecTree() {}
        AigLit operator() (Var tv, const VarSet& domain);
        void add_sample(const T& s) { samples.push_back(s); }
    protected:
        typedef unsigned long ulong;
        struct Split { std::vector<T> pos, neg; };
        const int verb;
        AigFactory& factory;
        std::vector<T> samples;
        void split_opp(const std::vector<T>& s, Var tv, Split& out);
        void split_me(Var d, const vector<DecTree::T>& src,
               std::vector<DecTree::T>& set0,
               std::vector<DecTree::T>& set1);
        void build(const Split& s, Var tv,
                VarSet& domain, Var max_domain_var, 
                vector<Lit>& stack, vector<SATSPC::LSet*>& trues, vector<SATSPC::LSet*>& falses);
        void calc_gains(const VarSet& domain, Var max_domain_var,  const Split& s,
                        std::vector<double>& gains);
        void calc_occs(const VarSet& domain,
                       const vector<T>& sset, std::vector<std::pair<ulong,ulong> >& occs);
        Var pick_decision(const VarSet& domain, Var max_domain_var,  const Split& s);
};
#endif /* DEC_TREE_H_20858 */
