%{
#include <iostream>
#include <string>
#include <map>
#include <stdio.h>
#include <string.h>
#include <unordered_map>
#include "auxiliary.h"
#include "aig.h"
#include "aig_util.h"
#include "qtypes.h"
using namespace std;
int yylex();
int qcir_line;  // incremented by the lexer
unordered_map<const char*,int,cstrHash,cstrEq> name2var;
unordered_map<int,const char*> var2name;
unordered_map<const char*,AigLit,cstrHash,cstrEq> name2id;
void yyerror(const char *m);
void start_qcir_lex(const string& filename); // in lexer
static string filename("<ERR>");

extern AigFactory factory;
AigUtil util(factory);
QAigFla qcir_qfla;
Var var_id(0);
static bool output_gate_polarity;
static char* output_gate_str;

#define TI(lit) AigLit::to_int(lit)
#define FI(num) AigLit::from_int(num)

std::vector<AigLit> qcir_AigLit_stack;
std::vector<Var> qcir_var_stack;

void semantic_error(string msg) {
  cerr<<filename<<":"<<qcir_line<<": "<<msg<<endl;
  exit(100);
}

int make_varid(const char* var_name) {
  const auto retv=++var_id;
  const auto i=name2var.insert(std::make_pair(var_name,retv));
  if (!i.second) semantic_error("var redefinition '"+string(var_name)+"'");
  var2name[retv] = var_name;
  return retv;
}

AigLit define_gate(const char *s, AigLit def) {
  const auto i=name2id.insert( std::pair<const char*,AigLit>(s,def) );
  if (!i.second) semantic_error("gate redefinition");
  return def;
}

AigLit get_AigLit(const char* s, bool polarity) {
 const auto i=name2id.find(s);
 if(i!=name2id.end()) {
   const AigLit retv=i->second;
   return polarity ? retv : factory.neg(retv);
  } else {
   const auto vi=name2var.find(s);
   if(vi==name2var.end()) semantic_error("undefined variable or gate '"+string(s)+"'");
   const Var v=vi->second;
   const AigLit retv=factory.mk_var(v,!polarity);
   return retv;
 }
}


void set_output_gate() {
  const auto i=name2id.find(output_gate_str);
  if (i==name2id.end()) semantic_error("output gate not defined");
  qcir_qfla.matrix=output_gate_polarity?i->second:factory.neg(i->second);
}

%}

//-- SYMBOL SEMANTIC VALUES -----------------------------
%defines
%name-prefix "qcir_"

%code requires {
}

%union {
  unsigned long long val;
  char* str;
}

%token <str> VAR_ID
%token FORMAT_ID NUM NL FORALL EXISTS XOR AND OR ITE OUTPUT
%token LP RP EQ SEMI PLUS MINUS COMMA
%token ERROR_TOKEN
%type <val> new_var
%type <val> lit
%type <val> quant
%type <val> gate_definition

//-- GRAMMAR RULES ---------------------------------------
%%

qcir_file: header qblocks output_stmt gates { set_output_gate(); }

nls: NL
   | nls NL
   ;

header : FORMAT_ID size nls { }
       ;

size:
    | NUM
    ;

qblocks:
       | qblocks qblock
       ;

qblock: quant LP var_list RP nls {
      const auto qt=($1)==0?EXISTENTIAL:UNIVERSAL;
      qcir_qfla.pref.push_back(Quantification(qt,VarVector(qcir_var_stack)));
      qcir_var_stack.clear();
      }
     ;

output_stmt : OUTPUT LP output_gate RP nls { };
output_gate: plus VAR_ID  {output_gate_str=$2;output_gate_polarity=true;}
           | MINUS VAR_ID {output_gate_str=$2;output_gate_polarity=false;}

gates:
  | gates gate
  ;

gate: VAR_ID EQ gate_definition {define_gate($1,FI($3));}
    ;

gate_definition:
      XOR LP lit COMMA lit RP nls {$$=TI(util.mk_xor(FI($3),FI($5)));}
    | AND LP lit_list RP nls      {$$=TI(util.and_(qcir_AigLit_stack, true)); qcir_AigLit_stack.clear();}
    | OR LP lit_list RP nls      {$$=TI(util.or_(qcir_AigLit_stack, true)); qcir_AigLit_stack.clear();}
    | ITE LP lit COMMA lit COMMA lit RP nls {$$=TI(util.ite(FI($3), FI($5), FI($7)));}
//    | quant LP var_list SEMI lit RP nls

quant: EXISTS  { $$=0; }
     | FORALL  { $$=1; }
     ;

var_list:
    | new_var                { qcir_var_stack.push_back($1); }
    | var_list COMMA new_var { qcir_var_stack.push_back($3); }
    ;

lit_list:
         | lit                  { qcir_AigLit_stack.push_back(FI($1)); }
         | lit_list COMMA lit   { qcir_AigLit_stack.push_back(FI($3)); }
         ;

plus:
    | PLUS
    ;

lit: plus VAR_ID   {$$=TI(get_AigLit($2,true));}
   | MINUS VAR_ID  {$$=TI(get_AigLit($2,false));}
    ;

new_var: VAR_ID { $$=make_varid($1); }

%%

void read_qcir(string _filename) {
  qcir_line=1;
  filename=_filename;
  start_qcir_lex(filename);
  qcir_parse();
}

void yyerror(const char *msg) {
  cerr<<filename<<":"<<qcir_line<<": "<<msg<<endl;
  exit(100);
}
