/*
 * File:  satsyn.cpp
 * Author:  mikolas
 * Created on:  25 Feb 2017 21:08:11
 * Copyright (C) 2017, Mikolas Janota
 */
#include"satsyn.h"
#include"aig_util.h"
#include"SeqCounter.h"
#include"minisat_auxiliary.h"
#include<stdio.h>
using std::unordered_map;
using std::vector;
using std::endl;
using std::cerr;
void satsyn::add_gate() {
    const auto sz = gates.size();
    gates.resize(sz+1);
    auto& g = gates[sz];
    add_lit(g.l1, sz);
    add_lit(g.l2, sz);
#if 0
    for (size_t i = 0; i < sz; ++i)
    for (size_t j = i; j < sz; ++j)
      solver.addClause(~mkLit(g.l1.enc_vars[i]),~mkLit(g.l2.enc_vars[j]));
#endif
}

void satsyn::add_lit(/*out*/lit_enc& l, size_t num_gates) {
    l.pol = newv();
    for(size_t i=0; i<input.size(); ++i)
        l.enc_vars.push_back(newv());
    for(size_t i=0; i < num_gates; ++i)
        l.enc_vars.push_back(newv());

#if 0
    DPRN("enc lit, pol"<<l.pol);
    for(size_t i=0; i<input.size(); ++i) std::cerr << " " << l.enc_vars[i];
    std::cerr<<std::endl;
    for(size_t i=0; i < num_gates; ++i) std::cerr << " " << l.enc_vars[input.size()+i];
    std::cerr<<std::endl;
#endif

    atmost1(l.enc_vars);
}

Lit satsyn::instantiate_gate(const gate_enc& g, const vec<lbool>& inputs, const vector<Lit>& gate_lits) {
    const auto l1 = instantiate_lit(g.l1, inputs, gate_lits);
    const auto l2 = instantiate_lit(g.l2, inputs, gate_lits);
    return enc_and(l1, l2);
}

Lit satsyn::instantiate_lit(const lit_enc& l, const vec<lbool>& inputs, const vector<Lit>& gate_lits) {
    vector<Lit> tmp;
    for(size_t vi = 0; vi < input.size(); ++vi) {
        const Var v = input[vi];
        const lbool val = eval(v, inputs);
        if (val == l_Undef) continue; //TODO: is this right?
        const Lit polLit = mkLit(l.pol, val == l_False);
        tmp.push_back(enc_and(mkLit(l.enc_vars[vi]), polLit));
    }
    const Lit polLit = mkLit(l.pol);
    for(size_t gi = 0; gi < gate_lits.size(); ++gi) {
        tmp.push_back(enc_and(mkLit(l.enc_vars[input.size()+gi]), iff(gate_lits[gi], polLit)));
    }
    return enc_or(tmp);
}


void satsyn::add_sample(bool positive, const vector<Lit>& input) {
    DPRN("sample "<<(positive ? "POS" : "NEG") <<" : "<<input);
    vec<lbool> vals;
    vals.growTo(solver.nVars(), l_Undef);
    for(Lit l : input) vals[var(l)] = sign(l) ? l_False : l_True;
    vector<Lit> gate_lits;
    for(gate_enc& g : gates) {
        gate_lits.push_back(instantiate_gate(g, vals, gate_lits));
    }
    const auto enc_out = instantiate_lit(out, vals, gate_lits);
    const auto gts = gates.size();
    if (gts) {
      solver.addClause(mkLit(out.enc_vars[input.size()+gts-1]));
    }
    solver.addClause(positive ? enc_out : ~enc_out);
}

void satsyn::atmost1(const vector<Var>& vs) {
    if (vs.size() < 5) {
        for (size_t i = 0; i < vs.size(); ++i)
        for (size_t j = i+1; j < vs.size(); ++j)
            solver.addClause(mkLit(vs[i], true), mkLit(vs[j], true));
    } else {
            SeqCounter s(1, solver, vs);
            s.encode();
    }
}

void satsyn::encode(uint num_gates) {
   for(uint i=0; i < num_gates; ++i) add_gate();
   add_lit(out, num_gates);
}

void satsyn::block(lit_enc& l, vec<Lit>& cl) {
  const auto& m = solver.model;
 // Var fv = -1;
  for(Var v : l.enc_vars) {
      cl.push(eval(v, m)==l_True ? ~mkLit(v) : mkLit(v));
/*
    if (eval(v, m)==l_True) {
      assert(fv==-1);
      fv = v;
    } */
  }
  cl.push(eval(l.pol, m)==l_True ? ~mkLit(l.pol) : mkLit(l.pol));
//  if (fv != -1) cl.push(~mkLit(fv));
}

void satsyn::block() {
    vec<Lit> cl;
    for(gate_enc& g : gates) {
      block(g.l1, cl);
      block(g.l2, cl);
    }
    //cerr<<"block:"<<cl<<endl;
    solver.addClause_(cl);
}

bool satsyn::run() {
   __TIMECODE(const bool r =  solver.solve(););
   return r;
}


AigLit satsyn::mk_gate(AigFactory& f, const gate_enc& g, const vector<AigLit>& gate2aig, const vec<lbool>& m) {
    const auto l1 = mk_lit(f, g.l1, gate2aig, m);
    const auto l2 = mk_lit(f, g.l2, gate2aig, m);
    return f.mk_and(l1, l2);
}

AigLit satsyn::_mk_lit(AigFactory& f, const lit_enc& l, const vector<AigLit>& gate2aig, const vec<lbool>& m) {
    Var fvi = -1;
    size_t vi;
    for(vi = 0; vi < input.size(); ++vi) {
        const Var encvar = l.enc_vars[vi];
        if (eval(encvar, m) != l_True) continue;
        fvi = vi;
        break;
    }
    const bool pol = eval(l.pol,m) == l_True;
    if (fvi >= 0) {
        return f.mk_var(input[fvi], !pol);
    } else {
        size_t gi;
        for(gi = 0; vi < l.enc_vars.size(); ++vi,++gi) {
            const Var encvar = l.enc_vars[vi];
            if (eval(encvar, m) != l_True) continue;
            fvi = vi;
            break;
        }
        if (fvi < 0) return f.mk_false();
        return pol ? gate2aig[gi]: f.neg(gate2aig[gi]);
    }
}

AigLit satsyn::mk_circ(AigFactory& f) {
    const auto& m = solver.model;
    //print_model(cerr << ":", m) << endl;
    vector<AigLit> gate2aig;
    for(gate_enc& g : gates) gate2aig.push_back(mk_gate(f, g, gate2aig, m));
    return mk_lit(f, out, gate2aig, m);
}
