/*
 * File:  satsyn.h
 * Author:  mikolas
 * Created on:  25 Feb 2017 21:08:31
 * Copyright (C) 2017, Mikolas Janota
 */
#ifndef SATSYN_H_9447
#define SATSYN_H_9447
#include"minisat_auxiliary.h"
#include"MiniSatExt.h"
#include"SatVarManager.h"
#include"aig.h"
#include"aig_util.h"
#include"LitSet.h"
#include"Encoder.h"
#include<vector>
#include<unordered_map>
using std::vector;
using SATSPC::MiniSatExt;
using SATSPC::Lit;
using SATSPC::Var;
class satsyn {
    public:
        satsyn(const std::vector<Var>& _input)
            : input(_input)
        {
            solver.new_variables(maxv(input));
        }
        void add_gate();
        void add_sample(bool positive, const vector<Lit>& input);
        void encode(uint num_gates);
        bool run();
        void block();
        AigLit mk_circ(AigFactory& f);
    protected:
        struct lit_enc {
            Var pol;
            vector<Var> enc_vars;
        };

        struct gate_enc { lit_enc l1, l2; };

        const std::vector<Var>&  input;

        MiniSatExt               solver;
        lit_enc                  out;
        std::vector<gate_enc>    gates;

        Lit enc_or(const vector<Lit>& ls) {
            if (ls.size()==1) return ls[0];
            const auto r = mkLit(newv());
            vec<Lit> cl;
            for (Lit l : ls) {
                cl.push(l);
                solver.addClause(~l, r);
            }
            cl.push(~r);
            solver.addClause_(cl);
            return r;
        }

        Lit enc_and(Lit l1, Lit l2) {
            vector<Lit> tmp(2);
            tmp[0]=l1;
            tmp[1]=l2;
            return enc_and(tmp);
        }

        Lit enc_and(const vector<Lit>& ls) {
            if (ls.size()==1) return ls[0];
            const auto r = mkLit(newv());
            vec<Lit> cl;
            for (Lit l : ls) {
                cl.push(~l);
                solver.addClause(l, ~r);
            }
            cl.push(r);
            solver.addClause_(cl);
            return r;
        }


        Lit iff(Lit l1, Lit l2) {
            const auto r = mkLit(newv());
            solver.addClause(~l1,~l2,r);
            solver.addClause(l1,l2,r);
            solver.addClause(l1,~l2,~r);
            solver.addClause(~l1,l2,~r);
            return r;
        }

        Lit iff(Var v1, Var v2) {
            return iff(mkLit(v1), mkLit(v2));
        }

        Var newv(){ return solver.newVar(); }
        void atmost1(const vector<Var>& vs);
        AigLit mk_gate(AigFactory& f, const gate_enc& g, const std::vector<AigLit>& gate2aig, const vec<lbool>& m);
        AigLit _mk_lit(AigFactory& f, const lit_enc& l, const std::vector<AigLit>& gate2aig, const vec<lbool>& m);
        AigLit mk_lit(AigFactory& f, const lit_enc& l, const std::vector<AigLit>& gate2aig, const vec<lbool>& m) {
            return _mk_lit(f, l, gate2aig, m);
            //const auto r = _mk_lit(f, l, gate2aig, m);
            //f.print(std::cerr<<"mk_lit: ",r)<<std::endl;
            //return r;
        }
        Lit instantiate_gate(const gate_enc& g, const vec<lbool>& inputs, const vector<Lit>& gate_lits);
        Lit instantiate_lit(const lit_enc& l, const vec<lbool>& inputs, const vector<Lit>& gate_lits);
        void add_lit(/*out*/lit_enc& l, size_t num_gates);
        void block(lit_enc& l, vec<Lit>& cl);
};

inline AigLit mkxor(AigFactory& f, AigLit l1, AigLit l2) {
    return f.mk_or(f.mk_and(l1, f.neg(l2)),f.mk_and(f.neg(l1), l2));
}

inline bool check_eq(AigFactory& f, Var maxv, AigLit a, AigLit b) {
    return false;
    const auto e = mkxor(f, a, b);
    MiniSatExt s;
    SatVarManager vm(s);
    aig2cnf enc(vm, 0, f, s);
    enc.alloc_vars(maxv);
    s.addClause(enc(e));
    std::cout << "Equality check: " << __FILE__ << ":" << __LINE__ << " " << std::endl;
    const double nn = read_cpu_time();
    const auto retv = s.solve();
    std::cout << "Equality check done: " << (read_cpu_time() - nn) << std::endl;
    std::cout<<(retv ? "different" : "same") << std::endl;
    return retv;
}

inline AigLit mkxor(const vector<Var>& vs, AigFactory& f) {
    auto retv = f.mk_false();
    for (Var v : vs) retv = mkxor(f, retv, f.mk_var(v, false));
    return retv;
}

inline bool test_satsyn5() {
    const Var N = 6;
    const uint S = 30;
    const uint G = 18;
    const uint M = 1000;
    vector<Var> vs;
    for(Var v = 1; v <= N; ++v) vs.push_back(v);
    std::cout<<"max gt: " << G <<std::endl;
    srand(0);
    satsyn s(vs);
    s.encode(G);
    for (uint k = 0; k < S; ++k)  {
      vector<Lit> t;
      bool p = false;
      for(Var v = 1; v <= N; ++v) {
        const bool rv = (rand()%100) < 50;
        t.push_back(rv ? mkLit(v) : ~mkLit(v));
        if (rv) p = !p;
      }
      s.add_sample(p, t);
      std::cout<<"sample count: " << (k+1) << ":\n" <<std::endl;
      std::cerr.flush();

      if (!s.run()) {
          std::cout<<"UNSAT\n";
          break;
      } else {
          AigFactory f;
          AigUtil u(f);
          const AigLit rc = s.mk_circ(f);
          const auto sz = u.size(rc);
          f.print(std::cout<<"res: (sz:"<<sz<<"):\n", rc, 1, true, true)<<std::endl;
          check_eq(f, N, rc, mkxor(vs, f));
          //f.print(std::cout<<"res: (sz:"<<sz<<") (gates:"<<gt<<")\n", rc, 3)<<std::endl;
      }
    }
    for (uint ver=0; ver<M; ++ver) {
        std::cout<<"version:"<<ver<<"\n";
        s.block();
        if (!s.run()) {
            std::cout<<"UNSAT\n";
            break;
        } else {
            AigFactory f;
            AigUtil u(f);
            const AigLit rc = s.mk_circ(f);
            const auto sz = u.size(rc);
            f.print(std::cout<<"res: (sz:"<<sz<<"):\n", rc, 1, true, true)<<std::endl;
            check_eq(f, N, rc, mkxor(vs, f));
            //f.print(std::cout<<"res: (sz:"<<sz<<") (gates:"<<gt<<")\n", rc, 3)<<std::endl;
        }
    }
    return true;
}

inline bool test_satsyn4() {
    vector<Var> vs(3);
    vs[0] = 1;
    vs[1] = 2;
    vs[2] = 3;
    satsyn s(vs);
    s.encode(2);
    vector<Lit> t(3);
    t[0] = mkLit(1);
    t[1] = mkLit(2);
    t[2] = mkLit(3);
    s.add_sample(true, t);

    t[0] = mkLit(1);
    t[1] = mkLit(2);
    t[2] = ~mkLit(3);
    s.add_sample(true, t);

    t[0] = ~mkLit(1);
    t[1] = ~mkLit(2);
    t[2] = mkLit(3);
    s.add_sample(true, t);

    t[0] = mkLit(1);
    t[1] = ~mkLit(2);
    t[2] = ~mkLit(3);
    s.add_sample(false, t);

    t[0] = ~mkLit(1);
    t[1] = mkLit(2);
    t[2] = mkLit(3);
    s.add_sample(true, t);

    t[0] = ~mkLit(1);
    t[1] = mkLit(2);
    t[2] = ~mkLit(3);
    s.add_sample(false, t);


    const bool r = s.run();
    assert(r);
    AigFactory f;
    const AigLit rc = s.mk_circ(f);
    f.print(std::cerr<<"res:\n", rc)<<std::endl;
    return r;
}


inline bool test_satsyn3() {
    vector<Var> vs(2);
    vs[0] = 1;
    vs[1] = 2;
    satsyn s(vs);
    s.encode(3);
    vector<Lit> t(2);
    t[0] = mkLit(1);
    t[1] = mkLit(2);
    s.add_sample(false, t);
    t[0] = mkLit(1);
    t[1] = ~mkLit(2);
    s.add_sample(true, t);
    t[0] = ~mkLit(1);
    t[1] = mkLit(2);
    s.add_sample(true, t);
    t[0] = ~mkLit(1);
    t[1] = mkLit(2);
    //s.add_sample(true, t);
    const bool r = s.run();
    assert(r);
    AigFactory f;
    const AigLit rc = s.mk_circ(f);
    f.print(std::cerr<<"res:\n", rc)<<std::endl;
    return r;
}


inline bool test_satsyn2() {
    vector<Var> vs(2);
    vs[0] = 1;
    vs[1] = 2;
    satsyn s(vs);
    s.encode(1);
    vector<Lit> t(2);
    t[0] = mkLit(1);
    t[1] = mkLit(2);
    s.add_sample(false, t);
    t[0] = mkLit(1);
    t[1] = ~mkLit(2);
    s.add_sample(true, t);
    const bool r = s.run();
    assert(r);
    AigFactory f;
    const AigLit rc = s.mk_circ(f);
    f.print(std::cerr<<"res:\n", rc)<<std::endl;
    AigUtil u(f);
    SATSPC::LSet a;
    a.insert(mkLit(1));
    a.insert(mkLit(2));
    u.eval(rc, Core(a,false)) == l_True;
    SATSPC::LSet a1;
    a1.insert(mkLit(1));
    a1.insert(~mkLit(2));
    u.eval(rc, Core(a1,false)) == l_False;
    return r;
}

inline bool test_satsyn1() {
    vector<Var> vs(2);
    vs[0] = 1;
    vs[1] = 2;
    satsyn s(vs);
    s.encode(0);
    vector<Lit> t(1);
    t[0] = mkLit(1);
    s.add_sample(true, t);
    t[0] = ~mkLit(1);
    s.add_sample(false, t);
    const bool r = s.run();
    assert(r);
    AigFactory f;
    const AigLit rc = s.mk_circ(f);
    f.print(std::cerr<<"res:\n", rc)<<std::endl;
    return r && rc.equals(f.mk_var(1,false));
}

inline bool test_satsyn() {
    return test_satsyn5();
//    return test_satsyn1() & test_satsyn2() & test_satsyn3() & test_satsyn4();
}
#endif
